@@ -24,7 +24,8 @@ class Inferences(object):
                 control_outcome, control_covariates, 
                 data,
                 placebo,
-                steps=8, verbose=False):
+                pen, steps=8, 
+                verbose=False):
         '''
         Solves the nested optimization function of finding the optimal synthetic control
 
@@ -41,7 +42,7 @@ class Inferences(object):
         '''
         args = (treated_outcome, treated_covariates,
                 control_outcome, control_covariates,
-                placebo, data)
+                pen, placebo, data)
         
         for step in range(steps):
 
@@ -50,13 +51,24 @@ class Inferences(object):
             #Subsequent times, sample a random pmf using the dirichlet distribution
             if step == 0:
                 v_0 = np.full(data.n_covariates, 1/data.n_covariates)
+                if pen == "auto":
+                    #if pen =="auto", we have an additional parameter to optimize over, so we append it
+                    v_0 = np.append(v_0, 0)
             else:
                 #Dirichlet distribution returns a valid pmf over n_covariates states
-                v_0 = np.random.dirichlet(np.ones(data.n_covariates), size=1)            
+                v_0 = np.random.dirichlet(np.ones(data.n_covariates), size=1)
+                if pen == "auto":
+                    #if pen =="auto", we have an additional parameter to optimize over, so we append it
+                    v_0 = np.append(v_0, np.random.lognormal(1.5, 1, size=1)) #Still experimenting with what distribution is appropriate
+            
 
             #Required to have non negative values
-            bnds = tuple((0,1) for _ in range(data.n_covariates))
-            
+            if pen != "auto":
+                bnds = tuple((0,1) for _ in range(data.n_covariates))
+            else:
+                #if pen =="auto", we have an additional parameter to optimize over, and we need to bound it to be non-negative
+                bnds = tuple((0,1000) if ((pen=="auto") and (x==data.n_covariates)) else (0,1) for x in range(data.n_covariates + 1))
+
             #Optimze
             res = minimize(self.total_loss, v_0,  args=(args),
                             method='L-BFGS-B', bounds=bnds, 
@@ -82,7 +94,7 @@ class Inferences(object):
     def total_loss(self, v_0, 
                     treated_outcome, treated_covariates,
                     control_outcome, control_covariates, 
-                    placebo, data):
+                    pen, placebo, data):
         '''
         Solves for w*(v) that minimizes loss function 1 given v,
         Returns loss from loss function 2 with w=w*(v)
@@ -96,15 +108,33 @@ class Inferences(object):
 
         n_controls = control_outcome.shape[1]
         
+        '''
         V = np.zeros(shape=(data.n_covariates, data.n_covariates))
         np.fill_diagonal(V, v_0)
+        '''
+        if pen == "auto":
+            V = np.diag(v_0[:-1])
+            pen_coef = v_0[-1]
+        else:
+            V = np.diag(v_0)
+            pen_coef = pen
         
         # Construct the problem - constrain weights to be non-negative
         w = cvx.Variable((n_controls, 1), nonneg=True)
         
         #Define the objective
-        objective = cvx.Minimize(cvx.sum(V @ cvx.square(treated_covariates - control_covariates @ w)))
+
+        #PROBLEM: treated_synth_difference = cvx.sum(V @ cvx.square(treated_covariates.T - control_covariates @ w)) runs better for normal sc,
+        #but it doesnt work at all for in-time placebos, this probably means I am messing up the dimensionality somewhere in the processing
+        #This is a work-around that works, but it ain't pretty
+        if placebo == 'in-time':
+            treated_synth_difference = cvx.sum(V @ cvx.square(treated_covariates - control_covariates @ w))
+        else:
+            treated_synth_difference = cvx.sum(V @ cvx.square(treated_covariates.T - control_covariates @ w))
         
+        pairwise_difference = cvx.sum(V @ (cvx.square(treated_covariates - control_covariates) @ w))
+        objective = cvx.Minimize(treated_synth_difference + pen_coef*pairwise_difference)
+
         #Add constraint sum of weights must equal one
         constraints = [cvx.sum(w) == 1]
         
@@ -122,10 +152,11 @@ class Inferences(object):
             if loss < data.min_loss:
                 data.min_loss = loss
                 data.w = w.value
-                data.v = v_0
+                data.v = np.diagonal(V)
+                data.pen = pen_coef
                 data.synth_outcome = data.w.T @ data.control_outcome_all.T #Transpose to make it (n_periods x 1)
                 data.synth_covariates = data.control_covariates @ data.w
-        
+
         elif placebo == "in-space":
             data.in_space_placebo_w = w.value
 
@@ -162,13 +193,18 @@ class Inferences(object):
             control_placebo_outcome = np.array([data.control_outcome_all[:,j] for j in range(data.n_controls) if j != i]).T
             control_placebo_covariates = np.array([[data.control_covariates[x,j] for j in range(data.n_controls) if j != i] for x in range(data.n_covariates)])
 
+            #Rescale covariates to be unit variance (helps with optimization)
+            treated_placebo_covariates, control_placebo_covariates = self._rescale_covariate_variance(treated_placebo_covariates,
+                                                                            control_placebo_covariates,
+                                                                            data.n_covariates)
+
             #Solve for best synthetic control weights
             self.optimize(treated_placebo_outcome[:data.periods_pre_treatment], 
                             treated_placebo_covariates,
                             control_placebo_outcome[:data.periods_pre_treatment], 
                             control_placebo_covariates,
                             data,
-                            "in-space", n_optim)
+                            "in-space", "auto", n_optim)
             
             #Compute outcome of best synthetic control
             if self.method == "SC":
@@ -221,12 +257,27 @@ class Inferences(object):
             placebo_treatment_period, data.treated_unit, data.n_controls, 
             data.periods_all, periods_pre_treatment, data.covariates
         )
-
+        '''
+        print("in_time_placebo_treated_covariates")
+        print(in_time_placebo_treated_covariates.shape,
+                "\n", in_time_placebo_treated_covariates)
+        print("in_time_placebo_control_covariates")
+        print(in_time_placebo_control_covariates.shape,
+                "\n", in_time_placebo_control_covariates)
+        '''
+        '''
+        #Rescale covariates to be unit variance (helps with optimization)
+        in_time_placebo_treated_covariates, in_time_placebo_control_covariates = self._rescale_covariate_variance(in_time_placebo_treated_covariates,
+                                                                            in_time_placebo_control_covariates,
+                                                                            data.n_covariates)
+        '''
+        #in_time_placebo_treated_covariates = in_time_placebo_treated_covariates.reshape(1, data.n_covariates)
         #Run find synthetic control from shortened pre-treatment period 
         self.optimize(in_time_placebo_treated_outcome, in_time_placebo_treated_covariates,
                         in_time_placebo_control_outcome, in_time_placebo_control_covariates,
                         data,
-                        "in-time", n_optim)
+                        "in-time", 0, n_optim)
+
         #Compute placebo outcomes using placebo_w vector from optimize
         placebo_outcome = data.in_time_placebo_w.T @ in_time_placebo_control_outcome_all.T
 
@@ -378,6 +429,7 @@ class Inferences(object):
         if not placebo:
             self.original_data.w = self.modified_data.w
             self.original_data.v = self.modified_data.v
+            self.original_data.pen = self.modified_data.pen
             self.original_data.synth_constant, self.original_data.synth_outcome = self._get_dsc_outcome(self.original_data.w,
                                                                                                         self.original_data.control_outcome_all,
                                                                                                         self.original_data.periods_pre_treatment,
