@@ -23,9 +23,10 @@ class SynthBase(object):
     
     def __init__(self, dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, control_units,
                 covariates, periods_all, periods_pre_treatment, n_controls, n_covariates,
-                treated_outcome, control_outcome, treated_covariates, control_covariates,
-                treated_outcome_all, control_outcome_all,
-                treatment_effect=None, w=None, **kwargs):
+                treated_outcome, control_outcome, treated_covariates, control_covariates, 
+                unscaled_treated_covariates, unscaled_control_covariates,
+                treated_outcome_all, control_outcome_all, pairwise_difference, pen,
+                w=None, v=None, **kwargs):
 
         '''
         INPUT VARIABLES:
@@ -59,8 +60,9 @@ class SynthBase(object):
         self.periods_all = periods_all
         self.periods_pre_treatment = periods_pre_treatment
         self.n_controls = n_controls
-        self.n_covariates = n_covariates 
-               
+        self.n_covariates = n_covariates
+        self.pen = pen
+
         
         '''
         PROCESSED VARIABLES:
@@ -104,14 +106,17 @@ class SynthBase(object):
         self.control_outcome = control_outcome
         self.treated_covariates = treated_covariates
         self.control_covariates = control_covariates
+        self.unscaled_treated_covariates = unscaled_treated_covariates
+        self.unscaled_control_covariates = unscaled_control_covariates
         self.treated_outcome_all = treated_outcome_all
         self.control_outcome_all = control_outcome_all
+        self.pairwise_difference = pairwise_difference
 
         ###Post inference quantities
-        self.w = w #Can be provided if using Synthetic DID
+        self.w = w #Can be provided
+        self.v = v #Can be provided
         self.weight_df = None
-        self.v = None
-        self.treatment_effect = treatment_effect #If known
+        self.comparison_df = None
         self.synth_outcome = None
         self.synth_constant = None
         self.synth_covariates = None
@@ -132,7 +137,7 @@ class SynthBase(object):
 
 class DataProcessor(object):
     
-    def _process_input_data(self, dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, **kwargs):
+    def _process_input_data(self, dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, pen, **kwargs):
         '''
         Extracts processed variables, excluding v and w, from input variables.
         These are all the data matrices.
@@ -152,23 +157,31 @@ class DataProcessor(object):
         control_units = dataset.loc[dataset[id_var] != treated_unit][id_var].unique()
 
         ###Get treated unit matrices first###
-        treated_outcome_all, treated_outcome, treated_covariates = self._process_treated_data(
+        treated_outcome_all, treated_outcome, unscaled_treated_covariates = self._process_treated_data(
             dataset, outcome_var, id_var, time_var, 
             treatment_period, treated_unit, periods_all, 
             periods_pre_treatment, covariates, n_covariates
         )
-        
+        '''
+        print("treated_covariates")
+        print(treated_covariates.shape,
+                "\n", treated_covariates)
+        '''
         ### Now for control unit matrices ###
-        control_outcome_all, control_outcome, control_covariates = self._process_control_data(
+        control_outcome_all, control_outcome, unscaled_control_covariates = self._process_control_data(
             dataset, outcome_var, id_var, time_var, 
             treatment_period, treated_unit, n_controls, 
             periods_all, periods_pre_treatment, covariates
         )
         
         #Rescale covariates to be unit variance (helps with optimization)
-        treated_covariates, control_covariates = self._rescale_covariate_variance(treated_covariates,
-                                                                            control_covariates,
+        treated_covariates, control_covariates = self._rescale_covariate_variance(unscaled_treated_covariates,
+                                                                            unscaled_control_covariates,
                                                                             n_covariates)
+        
+        #Get matrix of unitwise differences between control units to treated unit
+        pairwise_difference = self._get_pairwise_difference_matrix(treated_covariates, 
+                                                                   control_covariates)
 
         return {
             'dataset': dataset,
@@ -186,9 +199,13 @@ class DataProcessor(object):
             'treated_outcome_all': treated_outcome_all,
             'treated_outcome': treated_outcome,
             'treated_covariates': treated_covariates,
+            'unscaled_treated_covariates':unscaled_treated_covariates,
             'control_outcome_all': control_outcome_all,
             'control_outcome': control_outcome,
             'control_covariates': control_covariates,
+            'unscaled_control_covariates': unscaled_control_covariates,
+            'pairwise_difference':pairwise_difference,
+            'pen':pen
         }
     
     def _process_treated_data(self, dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, 
@@ -226,7 +243,7 @@ class DataProcessor(object):
         control_outcome = np.array(control_data[outcome_var]).reshape(n_controls, periods_pre_treatment).T
         
         #Extract the covariates for all the control units
-        #Identify which rows correspond to which control unit by setting index, 
+        #Identify which rows correspond to which control unit by setting index,
         #then take the unitwise mean of each covariate
         #This results in the desired (n_control x n_covariates) matrix
         control_covariates = np.array(control_data[covariates].\
@@ -244,15 +261,83 @@ class DataProcessor(object):
         big_dataframe /= np.apply_along_axis(np.std, 0, big_dataframe)
 
         #Re-seperate treated and control from big dataframe
-        treated_covariates = big_dataframe[:,0].reshape(1, n_covariates) #First column is treated unit
+        treated_covariates = big_dataframe[:,0].reshape(n_covariates, 1) #First column is treated unit
         control_covariates = big_dataframe[:,1:] #All other columns are control units
 
         #Return covariate matices with unit variance
         return treated_covariates, control_covariates
     
+    def _get_pairwise_difference_matrix(self, treated_covariates, control_covariates):
+        '''
+        Computes matrix of same shape as control_covariates, but with unit-wise difference from treated unit
+        Used in optimization objective for both SC and DSC
+        '''
+        return treated_covariates - control_covariates
+    
+    def _get_weight_df(self, data):
+      '''Prepares dataframe with weight assigned to each unit in synthetic control'''
+      
+      weight_df = pd.DataFrame({"Unit":data.control_units,
+                                        "Weight":data.w.T[0]})
+      #Show only units with non-zero weights
+      return weight_df.loc[weight_df["Weight"] > 0.00001] 
+    
+    def _get_comparison_df(self, data):
+      '''
+      Returns dataframe with shape (n_covariates, 4).
+      The four columns are:
+
+        self.original_data.treated_unit: 
+          Unscaled, average covariate values of the treated unit
+          If method == DSC, then the differenced data is displayed instead
+        
+        Synthetic Control: 
+          Unscaled, covariate values of the synthetic control unit
+          If method == DSC, then the differenced data is displayed instead
+      
+        WMAUE:
+          Weighted Mean Absolute Unitwise Error. For each covariate, how different is each control 
+          unit inside the synthetic control from the treated unit, weighted by the weight assigned to each unit.
+          This does not change even if method == DSC, as bias scales with value of difference and not change
+        
+        
+        Importance:
+          Leading diagonal of V matrix. How important, relative to other covariates,
+          is matching on each covariate in the optimization process?
+          Note that this is computed after rescaling each covariate to be unit variance, 
+          whereas the other columns show the unscaled covariate values.
+      
+      Interpretation:
+      If the synthetic control has good fit, the following things should be true:
+
+        1. Each row of the first two columns should be approximately equal. 
+           This means the synthetic control has reconstructed the treated unit values.
+
+        2. The third column should be small, relative to the values in columns 1 and 2. 
+           The closer to zero, the more similar the individual control units inside the syntetic control are to the treated unit.
+           The smaller the MWAUE, the lower the potential bias, all else equal.
+        
+        3. There is fixed way to interpret the importance column. Instead, it should be evaluated using domain knowledge.
+           Is the relative importance assigned to each covariate reasonable given the context?
+      '''
+      data = self.original_data if self.method=='SC' else self.modified_data
+
+      #WMAPE
+      unscaled_treated_covariates = self.original_data.unscaled_treated_covariates 
+      unscaled_control_covariates = self.original_data.unscaled_control_covariates
+      wmape = (np.abs((unscaled_treated_covariates - unscaled_control_covariates)) @ data.w).reshape(data.n_covariates,)
+      
+      comparison_df = pd.DataFrame({data.treated_unit: data.unscaled_treated_covariates.ravel(),
+                                    "Synthetic " + data.treated_unit: (data.unscaled_control_covariates @ data.w).ravel(),
+                                    "WMAPE": wmape,
+                                    "Importance":data.v},
+                                    index=data.covariates)
+      return comparison_df
+      
+
 class Synth(Inferences, Plot, DataProcessor):
 
-    def __init__(self, dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, n_optim=10, **kwargs):
+    def __init__(self, dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, n_optim=10, pen=0, **kwargs):
         '''
         data: 
           Type: Pandas dataframe. 
@@ -285,30 +370,37 @@ class Synth(Inferences, Plot, DataProcessor):
           Type: int. Default: 10. 
           Number of different initialization values for which the optimization is run. 
           Higher number means longer runtime, but a higher chance of a globally optimal solution.
+
+        pen:
+          Type: float. Default: 0.
+          Penalization coefficient which determines the relative importance of minimizing the sum of the pairwise difference of each individual
+          control unit in the synthetic control and the treated unit, vis-a-vis the difference between the synthtic control and the treated unit
+          Higher number means pairwise difference matters more.
         '''
         self.method = "SC"
 
         original_checked_input = self._process_input_data(
-            dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, **kwargs
+            dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, pen, **kwargs
         )
         self.original_data = SynthBase(**original_checked_input)
 
         #Get synthetic Control
         self.optimize(self.original_data.treated_outcome, self.original_data.treated_covariates,
                     self.original_data.control_outcome, self.original_data.control_covariates, 
-                    self.original_data, False, n_optim)
+                    self.original_data, False, pen, n_optim)
         
         #Compute rmspe_df
         self._pre_post_rmspe_ratios(None, False)
 
         #Prepare weight_df with unit weights
-        self.original_data.weight_df = pd.DataFrame({"Unit":self.original_data.control_units,
-                                        "Weight":self.original_data.w.T[0]})
+        self.original_data.weight_df = self._get_weight_df(self.original_data)
+        self.original_data.comparison_df = self._get_comparison_df(self.original_data)
+
 
 class DiffSynth(Inferences, Plot, DataProcessor):
 
     def __init__(self, dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, 
-                n_optim=10, not_diff_cols=None, **kwargs):
+                n_optim=10, pen=0, not_diff_cols=None, **kwargs):
         '''
         data: 
           Type: Pandas dataframe. 
@@ -352,28 +444,29 @@ class DiffSynth(Inferences, Plot, DataProcessor):
 
         #Process original data - will be used in plotting and summary
         original_checked_input = self._process_input_data(
-            dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, **kwargs
+            dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, pen, **kwargs
         )
         self.original_data = SynthBase(**original_checked_input)
 
         #Process differenced data - will be used in inference
         modified_dataset = self.difference_data(dataset, not_diff_cols)
         modified_checked_input = self._process_input_data(
-            modified_dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, **kwargs
+            modified_dataset, outcome_var, id_var, time_var, treatment_period, treated_unit, pen, **kwargs
         )
         self.modified_data = SynthBase(**modified_checked_input)
+        self.modified_data.pairwise_difference = self.original_data.pairwise_difference 
 
         #Get synthetic Control
         self.optimize(self.modified_data.treated_outcome, self.modified_data.treated_covariates,
                     self.modified_data.control_outcome, self.modified_data.control_covariates, 
-                    self.modified_data, False, n_optim)
+                    self.modified_data, False, pen, n_optim)
         
         #Compute rmspe_df for treated unit Synthetic Control
         self._pre_post_rmspe_ratios(None, False)
 
-        #Prepare weight_df with unit weights
-        self.weights_df = pd.DataFrame({"Unit":self.original_data.control_units,
-                                        "Weight":self.original_data.w.T[0]})
+        #Prepare tables
+        self.original_data.weight_df = self._get_weight_df(self.original_data)
+        self.original_data.comparison_df = self._get_comparison_df(self.original_data)
 
     def difference_data(self, dataset, not_diff_cols):
         '''
